from sam.managers import Manager
from sam.actions import Action
from swai_inst.exceptions import SwaiError
import swai_inst.installer as installer
import swai_inst.updater as updater
import swai_inst.utils as inst_utils
from typing import Optional, Tuple, List
import os
import yaml

class SwaiManager(Manager):
    title: str
    manager_id: str
    current_action = Action

    @property
    def manager_type(self) -> str:
        return "swai"

    # Packages will be formated like {app_id},{swai_url}
    # We need to parse the package_id and swai_url
    def parse_package_id(self, package_id: str) -> Tuple[str, Optional[str]]:
        parts = package_id.split(',')
        app_id = parts[0]
        swai_url = ','.join(parts[1:])        # Done incase url has commas
        if swai_url == "null":
            swai_url = None

        return app_id, swai_url
    
    def format_package_id(self, app_id: str, swai_url: Optional[str]) -> str:
        if swai_url is None:
            swai_url = "null"
        return f"{app_id},{swai_url}"
    
    def find_app_path_from_app_id(self, app_id: str) -> str:
        for swai_file in os.listdir("/usr/share/swai/apps"):
            swai_data = yaml.safe_load(f"/usr/share/swai/apps/{swai_file}")
            if swai_data.get("app_id") == app_id:
                return f"/usr/share/swai/apps/{swai_file}"
        return None

    def get_package_id_from_app_id(self, app_id: str) -> str:
        app_path = self.find_app_path_from_app_id(app_id)
        if app_path is not None:
            swai_data = yaml.safe_load(app_path)
            return self.format_package_id(app_id, swai_data.get("update_url"))
        return None

    def get_app_id(self, package_id: str) -> str:
        return package_id.split(',')[0]
    
    def get_swai_url(self, package_id: str) -> str:
        return package_id.split(',')[1]

    def install(self, action: Action):
        action.progress = 0
        try:
            _, swai_url = self.parse_package_id(action.package_id)
            installer.install_from_url(swai_url)
        except SwaiError as e:
            action.error = str(e)
        action.progress = 100
        action.status = "Installed"

    def remove(self, action: Action):
        action.progress = 0
        try:
            app_id, _ = self.parse_package_id(action.package_id)
            installer.remove_from_app_id(app_id)
        except SwaiError as e:
            action.error = str(e)
        action.progress = 99

    def update(self, action: Action):
        action.progress = 0
        try:
            app_id, _ = self.parse_package_id(action.package_id)
            updater.update_app_from_app_id(app_id)
        except SwaiError as e:
            action.error = str(e)
        action.progress = 99

    def check_update(self, package_id: str) -> bool:
        app_id, _ = self.parse_package_id(package_id)
        return updater.check_for_update(app_id)

    def check_updates(self) -> List[str]:
        updates = []
        # This function only returns the app ids
        # so we need to get the update urls
        update_app_ids = updater.check_for_updates()

        # Get update url for each app to put in the id
        for update in update_app_ids:
            update_package_id = self.get_package_id_from_app_id(update)
            if update_package_id is not None:
                updates.append(f"{update},{swai_url}")

        return updates

    def check_installed(self) -> List[str]:
        installed_apps = []
        for app in inst_utils.get_installed_apps():
            package_id = self.get_package_id_from_app_id(app.app_id)
            if package_id is not None:
                installed_apps.append(package_id)
        return installed_apps

    def bare_app_info(self, package_id: str) -> dict:
        app_id, _ = self.parse_package_id(package_id)
        app_path = self.find_app_path_from_app_id(app_id)
        if app_path is not None:
            swai_data = yaml.safe_load(app_path)
            return {
                "app_id": f"swai-{app_id.replace(".", "-")}",
                "name": swai_data.get("app_name"),
                "author": "Unknown",
                "primary_src": self.manager_id,
                "src_package_name": package_id,
                "icon_url": swai_data.get("icon_url"),
                "summary": swai_data.get("summary", "No summary available"),
                "description": swai_data.get("description", "No description available"),
            }

